/*****************************************************************************
*								18f4520_Flash.c								 *
******************************************************************************

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-5-8	Jonathan Karpick	Original version
2006-7-6	Barrett Heyneman	General code cleanup 

*****************************************************************************/

#include "18f4520_Flash.h"
#include "18fUtil.h"
#include <p18cxxx.h>
#include <string.h>


/*****************************************************************************
					PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/


/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/


/*****************************************************************************
						PRIVATE FUNCTION PROTOTYPES
*****************************************************************************/


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS
*****************************************************************************/

/******************************** WriteFlash *********************************

FLASH memory must be written in WRITE_LENGTH blocks.  If the address does not
lie on the start of a block, it is reset to the start of the block in which it
lies.  The number of block writes is computed (rounding up, based on 'bytes)
and those blocks of data are written to flash.  The last block may not be a
full block of data, and those other bytes are not changed in flash.  If the
data written will not completely fit in the data section of flash (as defined
in 18fUtil.h) nothing will be written and an appropriate error code will be
generated.

Last Modified:
Barrett Heyneman	2006-7-6

*****************************************************************************/

unsigned char WriteFlash(near unsigned short long addr, unsigned char *buff, unsigned int bytes)
{
	overlay unsigned int blocks;		//Number of blocks to write
	
		
	/* Reset the start adress to the start of the block. */
	addr &= ~(WRITE_LENGTH - 1);
	
	/* Make sure that the start is within the flash data section */
	if ((addr < FLASH_DATA) || (addr > FLASH_DATA_END))
	{
		return(FLASH_DATA_ACCESS);
	}
	
	/* Make sure that we're not going to try to write beyond the flash data
	 * section */
	if ((addr + bytes - 1) > FLASH_DATA_END)
	{
		return(FLASH_DATA_OVERRUN);
	}
	
	/* Get the number of block writes. */
	blocks = bytes/WRITE_LENGTH;
	if (bytes & (WRITE_LENGTH - 1))
	{
		blocks++;
	}
	
	/* Write all the blocks */
	do {
		/* Write the correct number of bytes into the holding registers */
		memcpyram2pgm(addr, buff, WRITE_LENGTH);
		
		/* Put the table pointer back into the block that we're trying to write. */
		TBLPTR--;
		
		/* Point to, access, and enable writes to FLASH program memory */
		EECON1bits.EEPGD = 1;
		EECON1bits.CFGS = 0;
		EECON1bits.WREN = 1;
		
		/* Need to disable interrupts during the write */
		INTCONbits.GIE = 0;
		
		/* Required sequence (not explained in datasheet */
		EECON2 = 0x55;
		EECON2 = 0xAA;
		
		/* Start the program (cpu stalls) */
		EECON1bits.WR = 1;
		
		/* Reenable interrupts */
		INTCONbits.GIE = 1;
		
		/* Disable writes to memory */
		EECON1bits.WREN = 0;
		
		/* Update the pointers to RAM and FLASH */
		addr += WRITE_LENGTH;
		buff += WRITE_LENGTH;
	} while (--blocks);
	
	return(SUCCESS);
}


/******************************** EraseFlash *********************************

Given the segment of memory to erase (from the start address and the length
'bytes'), all erase blocks which overlap that segment are erased.  If those
blocks contain other data, *IT WILL BE LOST*.  If the erase segment does not
completely lie in the data section of flash, nothing will be erased and the
appropriate error code will be generated.

Last Modified:
Barrett Heyneman	2006-7-6

*****************************************************************************/

unsigned char EraseFlash(near unsigned short long addr, unsigned int bytes)
{
	overlay unsigned int blocks = 0;		//Number of blocks
	
	
	/* Set the address to the start of the first block */
	addr &= ~(ERASE_LENGTH - 1);
	
	/* Make sure that the start is within the flash data section */
	if ((addr < FLASH_DATA) || (addr > FLASH_DATA_END))
	{
		return(FLASH_DATA_ACCESS);
	}
	
	/* Make sure that we're not going to try to write beyond the flash data
	 * section */
	if ((addr + bytes - 1) > FLASH_DATA_END)
	{
		return(FLASH_DATA_OVERRUN);
	}
	
	/* Get the number of erase blocks */
	blocks = bytes/ERASE_LENGTH;
	if (bytes & (ERASE_LENGTH - 1))
	{
		blocks++;
	}
	
	/* Erase all the blocks */
	do {
		TBLPTR = addr;
		
		/* Point to, access, and enable writes to FLASH program memory, and
		 * set for erasing. */
		EECON1bits.EEPGD = 1;
		EECON1bits.CFGS = 0;
		EECON1bits.WREN = 1;
		EECON1bits.FREE = 1;
		
		/* Need to disable interrupts during the write */
		INTCONbits.GIE = 0;
		
		/* Required sequence (not explained in datasheet */
		EECON2 = 0x55;
		EECON2 = 0xAA;
		
		/* Start the program (cpu stalls) */
		EECON1bits.WR = 1;
		
		/* Reenable interrupts */
		INTCONbits.GIE = 1;
		
		/* Disable writes to memory */
		EECON1bits.WREN = 0;
		
		/* Update the pointers to RAM and FLASH */
		addr += ERASE_LENGTH;
	} while (--blocks);
	
	return(SUCCESS);
}

unsigned char ReadFlashByte(near unsigned short long addr)
{
	//Set the address pointer
	TBLPTR = addr;

	//Read the data into RAM
	_asm
		//(Full text mnemonics must be used for table reads/writes)
		TBLRD
	_endasm

	//Return the data from the latch register
	return(TABLAT);
}

unsigned int ReadFlashTwoBytes(near unsigned short long addr)
{
	return( ((unsigned int)ReadFlashByte(addr)) +
			(((unsigned int)ReadFlashByte(addr + 1)) << 8) );
}

