#include "CtrlCloseLoop.h"
#include "ControllerMath.h"
#include "PhysicalConfig.h"
#include "servo.h"
#include "FileReader.h"
#include "18f4520_flash.h"
#include "StanceController.h"
#include "MatrixMath.h"
#include "Display.h"

static void take_step(void);
static void drain_adjustment(unsigned char drain);

static unsigned char ctrl_init = 0;
static unsigned char ctrl_enable = 0;

static unsigned int gait_location;
static unsigned int current_inx;
static unsigned int max_inx;

static unsigned char current_gait_matrix = 0xFF;

static ServoNumber_t current_servos;
//static ServoGroupNumber_t ff_ForceCmd;
static ServoGroupNumber_t fb_ForceCmd;
static SensorGroupNumber_t desired_force;
static ServoGroupNumber_t next_grasp_delta;
static ServoGroupNumber_t grasp_adjust; 

void CTRL_CL_EnableCtrl(void)
{
	if (ctrl_init)
		ctrl_enable = 1;
}

void CTRL_CL_DisableCtrl(void)
{
	ctrl_enable = 0;
}

void CTRL_CL_ResetCtrl(void)
{
	static unsigned char i;
	
	//Find the gait's location in memory
	gait_location = FR_FindEntry(CLOSE_LOOP_GAIT_FILE_ENTRY);

	//Check to make sure the gait was found
	if (gait_location != NULL_MEM_PTR)
	{
		//Read the size of the gait
		max_inx = ReadFlashTwoBytes(gait_location);

		//Advance the address by two bytes over the size of the gait
		gait_location += 2;
		
		//Make sure the gait isn't empty
		if (max_inx != 0)
		{
			//Initialize the gait
			current_inx = 0;
			
			//Zero out grasp adjustment vector
			for(i = 0; i < NUM_LEGS; i++)
				grasp_adjust.asValArray[i].raw = 0;
			
			//Load defauly grasp matrix (this can be removed once the dynamic gait_matrix loader is uncommented)
			LoadGraspMatrix(FOUR_FOOT_STANCE_CTRL);	

			//Take the first step
			take_step();
			
			//Set the initialization flag
			ctrl_init = 1;
		}
	}
	else
		//Indicate file error
		DSP_SetError(0x23);
}
	

void Ctrl_CloseLoop(void)
{
	//Check to make the gait is initialized and enabled
	if (ctrl_init)
	{
		if (ctrl_enable)
		{
			//Execute the loop
			take_step();
		}
		else
		{
			//Add current position and feed-forward force control
			SRV_SetAllAbs(&current_servos);
			//SRV_SetGroupRel(STROKE_SRV, &ff_ForceCmd);

			//Implement feedback force control
			StanceControl(&fb_ForceCmd);
			SRV_SetGroupRel(STROKE_SRV, &fb_ForceCmd);
			SRV_SetGroupRel(STROKE_SRV, &grasp_adjust);

			//Add any additional user offset and apply servo settings
			SRV_SetAllRel(SRV_GetOffset());
			SRV_UpdateServoValues();
		}
	}
	else
	{
		//Since the gait has not started, just apply the user offset
		SRV_SetAllAbs(SRV_GetOffset());
		SRV_UpdateServoValues();
	}
}

unsigned int Ctrl_CL_GetIndex(void)
{
	if (ctrl_init && ctrl_enable)
		return 0x1000 + current_inx;
	else
		return 0xFFFF;
}

static void take_step(void)
{
	static ServoType_t type;
	static Leg_t leg;
	static Number_t pos = 0;
	static unsigned int offset;
	static unsigned char new_gait_matrix = 0xFF;
	
	//Calculate the current position in the gait memory
	//offset = gait_location + (22 * current_inx);
	offset = gait_location + (20 * current_inx);
	
	//Loop through each servo and read it's position from memory
	for(type = 0; type < NUM_SERVO_TYPES; type++)
		for(leg = 0; leg < NUM_LEGS; leg++)
		{
			//Since the entire gait is precalculated we don't
			//need to worry about fractional bits
			pos.whole = ReadFlashByte(offset);
			
			//Update the sevo
			SRV_SetSingleAbs(leg, type, pos);
			
			//Advance one byte in memory
			offset++;
		}
	
	//Save position in memory
	SRV_GetAllTargets(&current_servos);

	//Set the desired forces from memory
	for(leg = 0; leg < NUM_LEGS; leg++)
	{
		//Read the desired force from memory
		desired_force.asValArray[leg].raw =
			ReadFlashTwoBytes(offset);
		
		//Advance two bytes in memory
		offset += 2;	
	}
	SetForceDesired(&desired_force);
	
	/*
	//Set correct Gait Matrix
	new_gait_matrix = ReadFlashByte(offset);
	offset++;
	if (new_gait_matrix != current_gait_matrix)
	{
		LoadGraspMatrix(new_gait_matrix);	
		current_gait_matrix = new_gait_matrix;
		
		StanceControl(&next_grasp_delta);
		VectorAdd(&grasp_adjust, &fb_ForceCmd, &grasp_adjust);
		VectorSub(&grasp_adjust, &next_grasp_delta, &grasp_adjust);
	}
	
	//If the gait indicates this a drain period (for example, during flight)
	//slowly remove any adjustments made during grasp matrix transistions
	drain_adjustment(ReadFlashByte(offset));
	offset++;
	*/
	
	//Perform feedforward force control
	//FeedForwardForce(&ff_ForceCmd);
	//SRV_SetGroupRel(STROKE_SRV, &ff_ForceCmd);
	
	//Perform feedback force control
	StanceControl(&fb_ForceCmd);
	SRV_SetGroupRel(STROKE_SRV, &fb_ForceCmd);
	SRV_SetGroupRel(STROKE_SRV, &grasp_adjust);
	
	//Add servo offset
	SRV_SetAllRel(SRV_GetOffset());
	
	//Send the updated values to the servos
	SRV_UpdateServoValues();

	//Advance one entry in memory
	current_inx++;
	//If the gait ended, loop back to the start
	if (current_inx >= max_inx)
		current_inx = 0;
}

#define DRAIN_CONSTANT 0x0280
static void drain_adjustment(unsigned char drain_flag)
{
	static Leg_t leg;
	static unsigned char flag_mask;
	
	flag_mask = 1;
	if (drain_flag != 0)
	{
		for(leg = 0; leg < NUM_LEGS; leg++)
		{
			if (drain_flag & flag_mask)
			{
			
				if (grasp_adjust.asValArray[leg].raw > 0)
				{
					grasp_adjust.asValArray[leg].raw =
						grasp_adjust.asValArray[leg].raw - DRAIN_CONSTANT;
				
					if (grasp_adjust.asValArray[leg].raw < 0)
						grasp_adjust.asValArray[leg].raw = 0;
				}
				else if (grasp_adjust.asValArray[leg].raw < 0)
				{
					grasp_adjust.asValArray[leg].raw =
						grasp_adjust.asValArray[leg].raw + DRAIN_CONSTANT;
				
					if (grasp_adjust.asValArray[leg].raw > 0)
						grasp_adjust.asValArray[leg].raw = 0;
				}
			}
			
			flag_mask <<= 1;
		}
	}
}
