/*****************************************************************************
 *                                 Display.c                                 *
 *****************************************************************************

Display routines for the StickyBot controller board.

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-6-30	Barrett Heyneman	Original version, abstracting out all display
									control (LEDs) and adding error display
									functionality.

*****************************************************************************/

#include "Display.h"
#include "Keys.h"
#include "serial.h"
#include <delays.h>
#include <p18cxxx.h>


/*****************************************************************************
					PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/
#define LED_TRIS			TRISC
#define LED_TRIS_VAL		0b11111100 //PORTC<1:0> outputs


/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/


/*****************************************************************************
						PRIVATE FUNCTION PROTOTYPES
*****************************************************************************/


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS
*****************************************************************************/

/****************************** InitDisplay **********************************

Initializes the pins for the two LED's to outputs and makes sure that they are
both turned off.

Last Modified:
Barrett Heyneman 	2006-6-30

*****************************************************************************/

void DSP_InitDisplay(void)
{
	/* Make the LED pins outputs */
	LED_TRIS &= LED_TRIS_VAL;
	
	/* Turn both LED's off */
	DSP_SetGreen(0);
	DSP_SetRed(0);
	
	return;
}


/********************************* Hello *************************************

Hello flashes the green LED, then the red, then the green again to indicate
that all modules were initialized correctly.

Last Modified:
Barrett Heyneman	2006-6-30

*****************************************************************************/

void DSP_Hello(void)
{
	/* Turn on the green LED */
	DSP_SetGreen(1);
	
	/* Wait for a total of 500ms */
	Delay10KTCYx(250);
	Delay10KTCYx(250);
	
	/* Turn the green off and the red on */
	DSP_SetGreen(0);
	DSP_SetRed(1);
	
	/* Wait for a total of 500ms */
	Delay10KTCYx(250);
	Delay10KTCYx(250);
	
	/* Turn the red off and the green on */
	DSP_SetRed(0);
	DSP_SetGreen(1);
	
	/* Wait for a total of 500ms */
	Delay10KTCYx(250);
	Delay10KTCYx(250);
	
	/* Turn off the green LED */
	DSP_SetGreen(0);
	
	return;
}


/********************************* SetError **********************************

Flashes the red LED a number of times corresponding to the error code sent as
the argument (consult Display.h for the error codes).  The upper nibble
indicates the module and is flashed quickly, while the lower nibble indicates
the particular error and is flashed slowly.  You cannot set a new error code
while another error is still being displayed; the function will return 0 if it
was unable to set the error code.

Last Modified:
Barrett Heyneman	2007-7-20

*****************************************************************************/

#define MODULE_MASK    0xF0
#define ERROR_MASK     0x0F
#define MODULE_TICKS   7
#define ERROR_TICKS    21

static unsigned char module_toggles_left = 0;
static unsigned char error_toggles_left = 0;
static unsigned char ticks_til_toggle = 0;
static unsigned char error_flag = 0;

unsigned char DSP_SetError(unsigned char err)
{	
	if ((module_toggles_left == 0) && (error_toggles_left == 0))
	{
		//Find the number of toggles for each (number of flashes x2)
		module_toggles_left = ((err & MODULE_MASK) >> 3);
		error_toggles_left = ((err & ERROR_MASK) << 1);
	
		//Start toggling on the next call to DSP_DisplayError
		ticks_til_toggle = 0;
		
		//Make sure the red LED starts turned off
		DSP_SetRed(0);
		
		//Set the error flag to indicate there is an error to display
		error_flag = 1;
		
		//Indicate that the error code was set correctly
		return(1);
	}
	
	//Indicate that the error code could not be set.
	return(0);
}


/****************************** DisplayError *********************************
 
Handles the actual toggling of the red LED for displaying error codes.  This
function must be called at a semi-regular interval for the error code to flash
at a consistent speed.

Last Modified:
Barrett Heyneman	2007-7-20

*****************************************************************************/
	
void DSP_DisplayError(void)
{	
	/* If we don't need to display an error, we can return right away */
	if (error_flag == 0)
	{
		return;
	}
	
	/* Otherwise, see if we need to toggle the LED again */
	if (ticks_til_toggle == 0)
	{
		/* First check module code toggles */
		if (module_toggles_left > 0)
		{
			module_toggles_left--;
			DSP_ToggleRed();
			ticks_til_toggle = MODULE_TICKS;
		}
		/* If the module code has been displayed, check the error code toggles */
		else if (error_toggles_left > 0 )
		{
			error_toggles_left--;
			DSP_ToggleRed();
			ticks_til_toggle = ERROR_TICKS;
		}
		else
		{
			/* If both toggles left were 0, we're done with the error code and can
		 	* clear the flag.
		 	*/
			error_flag = 0;
		}
	}
	else
		//Decrement the counter
		ticks_til_toggle--;
}


/********************************* SetHeartbeat ******************************

Sets the heartbeat postscalar.  The green LED will blink every hb_ticks calls
to DSP_Heartbeat, used primarily to indicate the board is still functioning.

Last Modified:
Barrett Heyneman	2007-7-20

*****************************************************************************/

static unsigned char heartbeat_bit = 0;
static unsigned char curr_hb_ticks = 0;


void DSP_SetHeartbeat(unsigned char hb_bit)
{
	if ((hb_bit >= 0) && (hb_bit < 8))
		heartbeat_bit = hb_bit;
}

/********************************* Heartbeat *********************************

Handles the actual display of the heartbeat, toggling the green LED every
2^(heartbeat_bit+1) calls.

Last Modified:
Barrett Heyneman	2007-7-20

*****************************************************************************/

void DSP_Heartbeat(void)
{
	if ((curr_hb_ticks >> heartbeat_bit) & 1)
	{
		DSP_ToggleGreen();
		curr_hb_ticks = 0;
		return;
	}
	curr_hb_ticks++;
}
