/*****************************************************************************
 *                                   Keys.c                                  *
 *****************************************************************************
 
Key press routines for StickyBot Controller board (specifically for PIC18Fxxx)
Assumes that the keys are normally high (pulled-up in hardware).

Date:		Author:				Comment:
------------------------------------------------------------------------------
2003-10-28	Virgilio Mattoli	Original (v1.3 assembly code)
2006-5-8	Jonathan Karpick	Port to C
2006-7-3	Barrett Heyneman	General code cleanup

*****************************************************************************/

#include "Keys.h"
#include "18fUtil.h"
#include <delays.h>
#include <p18cxxx.h>
#include "SBConfig.h"


/*****************************************************************************
						PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/

/* Hardware Settings */
#define KEY_PORT1				PORTE		// Keys are on Ports A,B, and E
#define KEY_PORT1_TRIS			TRISE
#define KEY_PORT1_TRIS_VAL		0b00000011	// Or with the TRIS to set the two key pins to inputs
#define KEY_PORT1_MASK			0b00000011
#define KEY_PORT1_SHIFT			0
#define KEY_PORT2				PORTA       
#define KEY_PORT2_TRIS			TRISA
#define KEY_PORT2_TRIS_VAL		0b00111111
#define KEY_PORT2_MASK			0b00110000
#define KEY_PORT2_SHIFT			0
#define KEY_PORT3				PORTB
#define KEY_PORT3_TRIS			TRISB
#define KEY_PORT3_TRIS_VAL		0b00110000
#define KEY_PORT3_MASK			0b00110000
#define KEY_PORT3_SHIFT			2
#define ADCON1_AD_VAL			0b00001111  // Value to or with ADCON1 to set all inputs to digital

/* Timing */
#define MAX_HOLD_TICKS			50			// Number of calls to CheckKeyEvent needed for a 'held' key
#define MIN_HOLD_TICKS			40			// Number of calls for a second held key (in a combo) to count


/* Type for storing information about particular keys in each bit */
typedef union {
	unsigned char byte;
	struct {
		unsigned B0:1;
		unsigned B1:1;
		unsigned R0:1;
		unsigned R1:1;
		unsigned R2:1;
		unsigned R3:1;
		unsigned NC:2;
	};
} KeyStatus_t;


/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/
/* Current state of the keys */
static KeyStatus_t curr_key_state;


/*****************************************************************************
						PRIVATE FUNCTION PROTOTYPES
*****************************************************************************/


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS	
*****************************************************************************/

/******************************** InitKeys ***********************************

Initializes the key pins as inputs

Last Modified
Barrett Heyneman	2006-7-3

*****************************************************************************/

void InitKeys(void)
{
	/* Use an '|=' because we are setting bits to 1's for input */
	KEY_PORT1_TRIS |= KEY_PORT1_TRIS_VAL;
	KEY_PORT2_TRIS |= KEY_PORT2_TRIS_VAL;
	KEY_PORT3_TRIS |= KEY_PORT3_TRIS_VAL;
	
	/* Set all the A/D pins to digital inputs (we don't use A/D anywhere) */
	ADCON1 |= ADCON1_AD_VAL;
	
	/* Give an initial value to curr_key_state */
	curr_key_state.byte = 0;
	
	return;
}


/****************************** CheckKeyEvent ********************************

Checks for any key events that have occured since the function was last called.
Single presses

Last Modified
Barrett Heyneman	2007-8-1

*****************************************************************************/

KeyEvent_t CheckKeyEvent(void)
{
	static KeyEvent_t event;
	static KeyStatus_t last_key_state;
	static KeyStatus_t keys_held = 0;
	static unsigned char key0_state = 0;
	static unsigned char key1_state = 0;
	static unsigned char B0_hold_time = 0;
	static unsigned char B1_hold_time = 0;
	static unsigned char R0_hold_time = 0;
	static unsigned char R1_hold_time = 0;
	static unsigned char R2_hold_time = 0;
	static unsigned char R3_hold_time = 0;
	static unsigned char wait_for_hold_release = 0;
	
	/* Save the last state and get the current state. */
	last_key_state = curr_key_state;
	curr_key_state.byte = 0;
	curr_key_state.byte |= ((KEY_PORT1 & KEY_PORT1_MASK) >> KEY_PORT1_SHIFT);
#ifdef WIRELESS_REMOTE
	curr_key_state.byte |= ((KEY_PORT2 & KEY_PORT2_MASK) >> KEY_PORT2_SHIFT);
	curr_key_state.byte |= ((KEY_PORT3 & KEY_PORT3_MASK) >> KEY_PORT3_SHIFT);
#endif
	
	/* Default of NO_EVENT */
	event = NO_EVENT;
	
	/* If we're waiting for a hold to release, check to see if it is finished. */
	if (wait_for_hold_release)
	{
		if (!(keys_held.byte & curr_key_state.byte))
		{
			/* No keys are being held anymore */
			wait_for_hold_release = 0;
			keys_held.byte = 0;
			
			/* Reset all the hold times */
			B0_hold_time = 0;
			B1_hold_time = 0;
			R0_hold_time = 0;
			R1_hold_time = 0;
			R2_hold_time = 0;
			R3_hold_time = 0;
		}
	}
	else
	{
		/* Check for falling edges on any of the keys; reset hold times and */
		if ((last_key_state.B0) && (!curr_key_state.B0))
		{
			event = B0_PRESS;
			B0_hold_time = 0;
			
		}
		else if ((last_key_state.B1) && (!curr_key_state.B1))
		{
			event = B1_PRESS;
			B1_hold_time = 0;
		}
		else if ((last_key_state.R0)&& (!curr_key_state.R0))
		{
			event = R0_PRESS;
			R0_hold_time = 0;
		}
		else if ((last_key_state.R1) && (!curr_key_state.R1))
		{
			event = R1_PRESS;
			R1_hold_time = 0;
		}
		else if ((last_key_state.R2) && (!curr_key_state.R2))
		{
			event = R2_PRESS;
			R2_hold_time = 0;
		}
		else if ((last_key_state.R3) && (!curr_key_state.R3))
		{
			event = R3_PRESS;
			R3_hold_time = 0;
		}
		
		/* Increment hold times for all the keys */
		if (curr_key_state.B0)
			B0_hold_time++;
		if (curr_key_state.B1)
			B1_hold_time++;
		if (curr_key_state.R0)
			R0_hold_time++;
		if (curr_key_state.R1)
			R1_hold_time++;
		if (curr_key_state.R2)
			R2_hold_time++;
		if (curr_key_state.R3)
			R3_hold_time++;
				
		/* Check to see if a key (or combo) has been held down */
		if (B0_hold_time > MAX_HOLD_TICKS)
		{
			keys_held.B0 = 1;
			if (B1_hold_time > MIN_HOLD_TICKS)
			{
				event = B0_B1_HOLD;
				keys_held.B1 = 1;
			}
			else
				event = B0_HOLD;	
			wait_for_hold_release = 1;
		}
		else if (B1_hold_time > MAX_HOLD_TICKS)
		{
			keys_held.B1 = 1;
			if (B0_hold_time > MIN_HOLD_TICKS)
			{
				event = B0_B1_HOLD;
				keys_held.B0 = 1;
			}
			else
				event = B1_HOLD;
			wait_for_hold_release = 1;
		}
		else if (R0_hold_time > MAX_HOLD_TICKS)
		{
			keys_held.R0 = 1;
			if (R1_hold_time > MIN_HOLD_TICKS)
			{
				event = R0_R1_HOLD;
				keys_held.R1 = 1;
			}
			else
				event = R0_HOLD;
			wait_for_hold_release = 1;
		}
		else if (R1_hold_time > MAX_HOLD_TICKS)
		{
			keys_held.R1 = 1;
			if (R0_hold_time > MIN_HOLD_TICKS)
			{
				event = R0_R1_HOLD;
				keys_held.R0 = 1;
			}
			else
				event = R1_HOLD;
			wait_for_hold_release = 1;
		}
		else if (R2_hold_time > MAX_HOLD_TICKS)
		{
			keys_held.R2 = 1;
			if (R3_hold_time > MIN_HOLD_TICKS)
			{
				event = R2_R3_HOLD;
				keys_held.R3 = 1;
			}
			else
				event = R2_HOLD;
			wait_for_hold_release = 1;
		}
		else if (R3_hold_time > MAX_HOLD_TICKS)
		{
			keys_held.R3 = 1;
			if (R2_hold_time > MIN_HOLD_TICKS)
			{
				event = R2_R3_HOLD;
				keys_held.R2 = 1;
			}
			else
				event = R3_HOLD;
			wait_for_hold_release = 1;
		}
	}
	
	return event;
}
