/*****************************************************************************
*                                 MatrixMath.c                               *
******************************************************************************

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-8-3	Barrett Heyneman	Original version
2006-8-9	Barrett Heyneman	Redid all functions to use new Number_t types
								found in ControllerMath.h
									
*****************************************************************************/

#include "MatrixMath.h"
#include "ControllerMath.h"


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS
*****************************************************************************/

/****************************** MatrixMatrix *********************************

Multiply two 4x4 matrices.  'M' must be in a distinct memory location from
'mat1' and 'mat2', otherwise operation is undefined.

*Limitation* In the interest of speed we don't caculate a NumberProduct_t,
then turn it into a Number_t, then calculate a NumberSum_t over all those
products, then turn it back into a Number_t (each time checking for overflow)
So, if the numbers in the matrix are big enough (or the matrices are big
enough) the matrix values returned could be bogus.

*NOTE* we may be able to speed this up more...but the typedef stuff seems
pretty fast.

M = mat1*mat2

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void MatrixMatrix(Number_t mat1[MATRIX_DIM][MATRIX_DIM],
                  Number_t mat2[MATRIX_DIM][MATRIX_DIM],
                  Number_t M[MATRIX_DIM][MATRIX_DIM])
{
	static NumberProduct_t temp;	//static will improve speed
	static unsigned char i;
	static unsigned char j;
	static unsigned char k;
	
	for (i = 0; i < MATRIX_DIM; i++)
	{
		for (j = 0; j < MATRIX_DIM; j++)
		{
			temp.raw = 0;
			for (k = 0; k < MATRIX_DIM; k++)
			{
				/* Compute the multiplication and keep adding it */
				temp.raw += NumMul(mat1[i][k], mat2[k][j]);
			}
			
			/* Grab the Number_t value from the product. */
			M[i][j] = temp.formatted.result;
		}
	}
	return;
}


/******************************** MatrixVector *******************************

Multiply a 4x4 matrix and a 4x1 vector.  'V' must be in a memory location which
is distinct from both 'mat' and 'vec', otherwise operation is undefined.

*Limitation* Same as for MatrixMatrix, if the values in the arguments are
large, or the dimension of the space is large enough, the vector values
returned could be bogus (overflow the type).

V = mat*vec

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void MatrixVector(Number_t mat[MATRIX_DIM][MATRIX_DIM],
                  Number_t vec[MATRIX_DIM], Number_t V[MATRIX_DIM])
{
	static NumberProduct_t temp;	// static will improve speed
	static unsigned char i;
	static unsigned char j;
	
	for (i = 0; i < MATRIX_DIM; i++)
	{
		temp.raw = 0;
		for (j = 0; j < MATRIX_DIM; j++)
		{
			/* Do the multiplication and keep adding. */
			temp.raw += NumMul(mat[i][j], vec[j]);
		}

		/* Grab the Number_t value from the product */
		V[i] = temp.formatted.result;
	}
	
	return;
}


/******************************* VectorDot ***********************************

Compute the dot product of two vectors.

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

Number_t VectorDot(Number_t vec1[MATRIX_DIM], Number_t vec2[MATRIX_DIM])
{
	static NumberProduct_t temp;	// static will improve speed
	static unsigned char i;
	
	temp.raw = 0;
	for (i = 0; i < MATRIX_DIM; i++)
	{
		/* Do the multiplication and keep summing */
		temp.raw += NumMul(vec1[i], vec2[i]);
	}
	
	/* Grab the Number_t value from the product */
	return(temp.formatted.result);
}


/******************************* VectorAdd ***********************************

Adds two Nx1 vectors.


Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void VectorAdd(Number_t vec1[MATRIX_DIM], Number_t vec2[MATRIX_DIM],
               Number_t V[MATRIX_DIM])
{
	static NumberSum_t temp;
	static unsigned char i;
	
	for (i = 0; i < MATRIX_DIM; i++)
	{
		temp.raw = NumAdd(vec1[i], vec2[i]);
		V[i] = temp.formatted.result;
	}
	
	return;
}


/***************************** VectorSub *************************************

Subtracts vector 2 from vector 1.

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void VectorSub(Number_t vec1[MATRIX_DIM], Number_t vec2[MATRIX_DIM],
               Number_t V[MATRIX_DIM])
{
	static NumberSum_t temp;
	static unsigned char i;
	
	for (i = 0; i < MATRIX_DIM; i++)
	{
		temp.raw = NumSub(vec1[i], vec2[i]);
		V[i] = temp.formatted.result;
	}
	return;
}


/******************************* VectorMul ***********************************

Multiplies vectors element by element.

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void VectorMul(Number_t vec1[MATRIX_DIM], Number_t vec2[MATRIX_DIM],
               Number_t V[MATRIX_DIM])
{
	static NumberProduct_t temp;
	static unsigned char i;
	
	for (i = 0; i < MATRIX_DIM; i++)
	{
		temp.raw = NumMul(vec1[i], vec2[i]);
		V[i] = temp.formatted.result;
	}
	
	return;
}

/******************************* VectorCpy ***********************************

Copy a vector from one location in memory to another

Last Modified:
Salomon Trujillo	2006-11-10

*****************************************************************************/

void VectorCpy(Number_t source[MATRIX_DIM], Number_t dest[MATRIX_DIM])
{
	static unsigned char i;
	for (i = 0; i < MATRIX_DIM; i++)
		dest[i] = source[i];
}

/******************************* VectorSclrMul ***********************************

Multiplies a vector using a scalar

Last Modified:
Salomon Trujillo	2006-11-15

*****************************************************************************/

void VectorSclrMul(Number_t vec1[MATRIX_DIM], Number_t scalar,
                   Number_t V[MATRIX_DIM])
{
	static NumberProduct_t temp;
	static unsigned char i;
	
	for (i = 0; i < MATRIX_DIM; i++)
	{
		temp.raw = NumMul(vec1[i], scalar);
		V[i] = temp.formatted.result;
	}
	
	return;
}
