/*****************************************************************************
*                                PhysicalConfig.h                            *
******************************************************************************

This file will hold the complex types for both sensor and servo data, as well
as defined constants for accessing certain sensor/servo types and legs.  We
should keep this all in one spot so that it is easy to check and make sure that
all the servo/sensor relations are correct if/when we change any physical
connections on StickyBot

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-8-9	Barrett Heyneman	First Draft

*****************************************************************************/

#ifndef __PHYSICAL_CONFIG_H
#define __PHYSICAL_CONFIG_H

#include "ControllerMath.h"


/******	General Parameters and Types ****************************************/

#define NUM_LEGS			4


typedef enum {
	FRONT_LEFT = 0,
	FRONT_RIGHT = 1,
	BACK_LEFT = 2,
	BACK_RIGHT = 3,
	NO_LEG = 255
} Leg_t;


/******	Sensor Parameters and Types *****************************************/

#define NUM_SENSOR_TYPES	1
#define NUM_SENSORS			(NUM_LEGS * NUM_SENSOR_TYPES)

typedef enum {
	STROKE_SNS = 0,
	NO_SNS = 255
} SensorType_t;

/* Type which holds all of some unsigned char data for a single sensor type
 * (such as raw sensor readings).  The values can be accessed as an array
 * (indexed by leg number) or directly by leg name.
 */
typedef union {
	unsigned char asValArray[NUM_LEGS];
	struct {
		unsigned char frontLeft;
		unsigned char frontRight;
		unsigned char backLeft;
		unsigned char backRight;
	} byLeg;
} SensorGroupRaw_t; //RawTypeValues_t;

/* Type which holds all of some Number_t data for a single sensor type (such
 * as calibrated sensor force readings). The values can be accessed as an
 * array (indexed by leg number) or directly by leg name.
 */
typedef union {
	Number_t asValArray[NUM_LEGS];
	struct {
		Number_t frontLeft;
		Number_t frontRight;
		Number_t backLeft;
		Number_t backRight;
	} byLeg;
	struct {
		Number_t frontLeft;
		Number_t frontRight;
		Number_t backLeft;
		Number_t backRight;
	};
} SensorGroupNumber_t; //ForceTypeValues_t

/* Type to hold all of some unsigned char data for all sensors.  The data can
 * be accessed as:
 *		A table of unsigned chars indexed by [type][leg]
 *		An array of SensorGroupRaw_t indexed by [type]
 *		A single SensorGroupRaw_t object by type name (ie. X.byType.stroke)
 *      A single unsigned char by full name (ie. X.byName.FR_Stroke)
 */
typedef union {
	/* Table index: [type #][leg #] */
	unsigned char asTable[NUM_SENSOR_TYPES][NUM_LEGS];
	/* Type array index: [type #] */
	SensorGroupRaw_t asTypeArray[NUM_SENSOR_TYPES];
	struct {
		SensorGroupRaw_t stroke;
	} byType;
	struct {
		unsigned char FL_Stroke;
		unsigned char FR_Stroke;
		unsigned char BL_Stroke;
		unsigned char BR_Stroke;
	} byName;
} SensorRaw_t;


/* Type to hold all of some Number_t data for all the sensors.  The data can
 * be accessed as:
 *		A table of Number_t objects indexed by [type][leg]
 *		An array of SensorGroupNumber_t indexed by [type]
 *		A single SensorGroupNumber_t object by type name (ie. X.byType.stroke)
 *      A single Number_t object by full name (ie. X.byName.FR_Stroke)
 */
typedef union {
	/* Table index: [type #][leg #] */
	Number_t asTable[NUM_SENSOR_TYPES][NUM_LEGS];
	/* Type array index: [type #] */
	SensorGroupNumber_t asTypeArray[NUM_SENSOR_TYPES];
	struct {
		SensorGroupNumber_t stroke;
	} byType;
	struct {
		Number_t FL_Stroke;
		Number_t FR_Stroke;
		Number_t BL_Stroke;
		Number_t BR_Stroke;
	} byName;
	struct {
		Number_t FL_Stroke;
		Number_t FR_Stroke;
		Number_t BL_Stroke;
		Number_t BR_Stroke;
	};
} SensorNumber_t;


/******	Servo Parameters and Types ******************************************/

#define NUM_SERVO_TYPES		3
#define NUM_SERVOS			12 /*(NUM_LEGS * NUM_SERVO_TYPES)*/

typedef enum {
	FOOT_SRV = 0,
	STROKE_SRV = 1,
	WING_SRV = 2,
	NO_SRV = 255
} ServoType_t;


/* Type which holds all of some unsigned char data for a single servo type
 * (such as the actual servo commands).  The values can be accessed as an
 * array (indexed by leg number) or directly by leg name.
 */
typedef union {
	unsigned char asValArray[NUM_LEGS];
	struct {
		unsigned char frontLeft;
		unsigned char frontRight;
		unsigned char backLeft;
		unsigned char backRight;
	} byLeg;
} ServoGroupRaw_t; //RawTypePositions_t;

/* Type which holds all of some Number_t data for a single servo type (such as
 * precise relative servo positions). The data can be accessed as an array
 * (indexed by leg number) or directly by leg name.
 */
typedef union {
	Number_t asValArray[NUM_LEGS];
	struct {
		Number_t frontLeft;
		Number_t frontRight;
		Number_t backLeft;
		Number_t backRight;
	} byLeg;
} ServoGroupNumber_t; //ExactTypePositions_t;

/* Type to hold all of some unsigned char data for all the servos.  The data
 * can be accessed as:
 *		A table of unsigned chars indexed by [type][leg]
 *		An array of ServoGroupRaw_t indexed by [type]
 *		A single ServoGroupRaw_t object by type name (ie. X.byType.foot)
 *      A single unsigned chars by full name (ie. X.byName.BL_Wing)
 */
typedef union {
	/* Table index: [type #][leg #] */
	unsigned char asTable[NUM_SERVO_TYPES][NUM_LEGS];
	unsigned char asArray[NUM_SERVOS];
	/* Type array index: [type #] */
	ServoGroupRaw_t asTypeArray[NUM_SERVO_TYPES];
	struct {
		ServoGroupRaw_t foot;
		ServoGroupRaw_t stroke;
		ServoGroupRaw_t wing;
	} byType;
	struct {
		unsigned char FL_Foot;
		unsigned char FR_Foot;
		unsigned char BL_Foot;
		unsigned char BR_Foot;
		unsigned char FL_Stroke;
		unsigned char FR_Stroke;
		unsigned char BL_Stroke;
		unsigned char BR_Stroke;
		unsigned char FL_Wing;
		unsigned char FR_Wing;
		unsigned char BL_Wing;
		unsigned char BR_Wing;
	} byName;
} ServoRaw_t; //RawPositions_t;


/* Type to hold all of some Number_t data for all the servos.  The data can
 * be accessed as:
 *		A table of Number_t objects indexed by [type][leg]
 *		An array of ServoGroupNumber_t indexed by [type]
 *		A single ServoGroupNumber_t object by type name (ie. X.byType.wing)
 *      A single Number_t object by full name (ie. X.byName.BR_Foot)
 */
typedef union {
	/* Raw values */
	signed int raw[NUM_SERVOS];
	/* Table index: [type #][leg #] */
	Number_t asTable[NUM_SERVO_TYPES][NUM_LEGS];
	/* Type array index: [type #] */
	ServoGroupNumber_t asTypeArray[NUM_SERVO_TYPES];
	struct {
		ServoGroupNumber_t foot;
		ServoGroupNumber_t stroke;
		ServoGroupNumber_t wing;
	} byType;
	struct {
		ServoGroupNumber_t foot;
		ServoGroupNumber_t stroke;
		ServoGroupNumber_t wing;
	};
	struct {
		Number_t FL_Foot;
		Number_t FR_Foot;
		Number_t BL_Foot;
		Number_t BR_Foot;
		Number_t FL_Stroke;
		Number_t FR_Stroke;
		Number_t BL_Stroke;
		Number_t BR_Stroke;
		Number_t FL_Wing;
		Number_t FR_Wing;
		Number_t BL_Wing;
		Number_t BR_Wing;
	} byName;
	struct {
		Number_t FL_Foot;
		Number_t FR_Foot;
		Number_t BL_Foot;
		Number_t BR_Foot;
		Number_t FL_Stroke;
		Number_t FR_Stroke;
		Number_t BL_Stroke;
		Number_t BR_Stroke;
		Number_t FL_Wing;
		Number_t FR_Wing;
		Number_t BL_Wing;
		Number_t BR_Wing;
	};
	struct {
		signed int FL_Foot;
		signed int FR_Foot;
		signed int BL_Foot;
		signed int BR_Foot;
		signed int FL_Stroke;
		signed int FR_Stroke;
		signed int BL_Stroke;
		signed int BR_Stroke;
		signed int FL_Wing;
		signed int FR_Wing;
		signed int BL_Wing;
		signed int BR_Wing;
	}raw_byName;
} ServoNumber_t; //ForceValues_t;



#endif /* __PHYSICAL_CONFIG_H */
