/*****************************************************************************
*                           StanceController.c                               *
******************************************************************************

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-8-1	Barrett Heyneman	Original Version
2006-8-10	Barrett Heyneman	Reworked to take advantage of the new Number,
								Sensor, and Servo data types.

*****************************************************************************/

#include "StanceController.h"
#include "18fSensors.h"
#include "18fUtil.h"
#include "18f4520_flash.h"
#include "MatrixMath.h"
#include "ControllerMath.h"
#include "PhysicalConfig.h"
#include "FileReader.h"

/*****************************************************************************
						PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/


/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/

/* The grasp matrix and its inverse */
static Number_t Grasp[4][4];
static Number_t InvGrasp[4][4];

/* The grasp-space gains */
static Number_t Gains[4];

/* The pseduo-inverse of the grasp-space gains */
static Number_t InvGains[4];

/* The desired forces */
static Number_t FDes[4];

static unsigned char grasp_init = 0;


// (hard code for now, but load from flash later.)
static SensorGroupNumber_t ff_stiffness_gains = {0xFF29, 0xFECF, 0xFED1, 0xFF3C};


/*****************************************************************************
						PRIVATE FUNCTION DEFINITIONS
*****************************************************************************/


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS
*****************************************************************************/


/**************************** InitStanceControl ******************************

Reads in the grasp matrix and its inverse, the grasp space gains, desired
grasp space forces, and the initial grasp space feed-forward velocity from
flash and loads them in local variables.

Last Modified:
Barrett Heyneman	2006-8-2

*****************************************************************************/
 
void LoadGraspMatrix(unsigned char GraspId)
{
	near unsigned short long read_loc;
	
	read_loc = (near unsigned short long)FR_FindEntry(GraspId);
	if (read_loc == NO_ENTRY_FOUND)
		/* Change to an error code */
		return;
		
	/* Grab all the data */
	
	/* First 16 bytes of data are the Grasp Matrix */
	ReadFlash(read_loc, &Grasp, 16*NUMBER_SIZE);
	read_loc += 16*NUMBER_SIZE;
	
	/* Next 16 are the inverse grasp matrix */
	ReadFlash(read_loc, &InvGrasp, 16*NUMBER_SIZE);
	read_loc += 16*NUMBER_SIZE;
	
	/* Next 4 are the gains */
	ReadFlash(read_loc, &Gains, 4*NUMBER_SIZE);
	read_loc += 4*NUMBER_SIZE;
	
	//See initialization flag
	grasp_init = 1;

	return;
}


/****************************** StanceControl ********************************

Applies the stance control law to all the legs and puts the computed servo
deltas into memory starting at address 'deltas' (just the stroke servos)

At last check, this controller code took 648 us to execute.

Last Modified:
Barrett Heyneman	2006-8-29

*****************************************************************************/

void StanceControl(ServoGroupNumber_t *stroke_deltas)
{
	static SensorGroupNumber_t forces;
	static Number_t temp[NUM_LEGS];
	static Number_t temp2[NUM_LEGS];
	
	//Check for initialization
	if (!grasp_init)
		return;

	/* Read in the stroke motor forces. */
	SNS_GetTypeForce(STROKE_SNS, &forces);
	
	/* Find the difference between actual and desired grasp space forces */
	VectorSub(forces.asValArray, FDes, temp);

	/* Convert to grasp space by left multiplying by the grasp matrix */
	MatrixVector(Grasp, temp, temp2);
	
	/* Multiply that by the diagonal gain matrix (represented as a vector) */
	VectorMul(temp2, Gains, temp);
	
	/* Multiply by the inverse grasp matrix to get deltas */
	MatrixVector(InvGrasp, temp, stroke_deltas);
	
	return;
}

void SetForceDesired(SensorGroupNumber_t *force_desired)
{
	static unsigned leg;

	for(leg = 0; leg < NUM_LEGS; leg++)
		FDes[leg] = force_desired->asValArray[leg];
}

void GetForceDesired(SensorGroupNumber_t *force_desired)
{
	static unsigned leg;

	for(leg = 0; leg < NUM_LEGS; leg++)
		force_desired->asValArray[leg] = FDes[leg];
}

void SetSingleForceDesired(Leg_t leg, Number_t force_desired)
{
	FDes[leg] = force_desired;
}

Number_t GetSingleForceDesired(Leg_t leg)
{
	return(FDes[leg]);
}


void FeedForwardForce(ServoGroupNumber_t *deltas)
{
	VectorMul(FDes, &ff_stiffness_gains, deltas);
	//VectorSclrMul(deltas, invert, deltas);
}
