#include <p18cxxx.h>
#include <timers.h>
#include "servo.h"
#include "serial.h"
#include "display.h"
#include "interrupts.h"

//Capture compare period
//(20 ms * 40 MHz) / (4 * 8) = 25000 = 0x61A8
#define CMP_HL 0x61A8

/* These registers are used by the delay functions. */
extern unsigned char DelayCounter1;
extern unsigned char DelayCounter2;

//Private variable for timer event
unsigned char TimerEventFlag = 0;

/* Private variables to save the counts for delays */
static unsigned char DelayTemp1;
static unsigned char DelayTemp2;

//Private function prototypes
void highISR(void);
static void TimerUpdate(void);

////////////////////////////////
// HIGH INTERRUPT SERVICE ROUTINE
////////////////////////////////

// Code Vectoring
#pragma code high_vector=0x08
void interrupt_at_high_vector(void)
{
_asm GOTO highISR _endasm
}
#pragma code /* return to the default code section */

// High ISR Code
#pragma interruptlow highISR
void highISR( void )
{
	/* Save the counter values; need to because the servo PWM generation uses
	 * the delay counters and may overwrite counter values if we interrupt
	 * during a call to one of the delay functions.
	 */
	DelayTemp1 = DelayCounter1;
	DelayTemp2 = DelayCounter2;
	
	if (PIR2bits.CCP2IF)
		TimerUpdate();

	//Flag checking and clearing is done automatically in this function
	SRL_HandleSerialPort();	
	
	/* Restore the original delay counts */
	DelayCounter1 = DelayTemp1;
	DelayCounter2 = DelayTemp2;
}

////////////////////////////////
// LOW INTERRUPT SERVICE ROUTINE
////////////////////////////////

// Code Vectoring
#pragma code low_vector=0x18
void interrupt_at_low_vector(void)
{
_asm GOTO highISR _endasm
}
#pragma code /* return to the default code section */

static void TimerUpdate( void )	// must be less than 20ms (including interrupt overhead)
{
	PIR2bits.CCP2IF = 0;	// Clear compare interrupt flag
	SRV_GenerateServoPWM();		// Generate the servo PWM signal.
	TimerEventFlag = 1;		// Set timer flag
}

unsigned char HasTimerTriggered(void)
{
	//Check flag, but clear it after reading it.
	if (TimerEventFlag)
	{
		TimerEventFlag = 0;
		return(1);
	}	
	return(0);
}

void RedLEDifTimerTriggered(void)
{
	if (TimerEventFlag)
		DSP_SetRed(1);
}

void INRPT_InitTimer(void)
{
	INRPT_DisableTimer();
	
	//Initialize Timer
	CCPR2 = CMP_HL;
	CCP2CON = 0b00001011;
	OpenTimer3(TIMER_INT_ON & T3_8BIT_RW & T3_SOURCE_INT & T3_PS_1_8 &
				T1_CCP1_T3_CCP2);

	INRPT_DisableTimer();
}

void INRPT_EnableTimer(void)
{
	//Reset the timer
	TMR3L = 0;
	TMR3H = 0;

	//Clear any flags
	PIR2bits.CCP2IF = 0;

	//Enable the timer interrupt
	PIE2bits.CCP2IE = 1;

	//Enable the timer
	T3CONbits.TMR3ON = 1;
}

void INRPT_DisableTimer(void)
{
	//Clear any flags
	PIR2bits.CCP2IF = 0;

	//Disable the timer interrupt
	PIE2bits.CCP2IE = 0;

	//Disable the timer
	T3CONbits.TMR3ON = 0;

	//Clear any flags accidentally set
	PIR2bits.CCP2IF = 0;
}
