/*****************************************************************************
 *                                 serial.c                                  *
 *****************************************************************************
Author: Salomon Trujillo
 
Functions to provide a serial interface, as well as servo command control,
flash downloading and uploading.

The serial code is interrupt driven, both for transmit and receive.

Transmission can be queued by using SendCmd or SendData.
SendCmd will send the data in raw format.
SendData will escape all bytes that match current command bytes.

*******************************/
#define SRL_SRV_CMD_SND_TO_CTRL

//Includes
#include <p18cxxx.h>	//MPLAB converts this to the correct processor header
#include "18fUtil.h"
#include "18f4520_flash.h"
#include "serial.h"
#include "servo.h"
#include "18fSensors.h"
#include "PhysicalConfig.h"
#include "CtrlCloseLoop.h"
#include "CtrlOpenLoop.h"
#include "Display.h"
#include "StanceController.h"

/************* Constants *************/
#define SRL_SUCCESSFUL 1
#define SRL_UNSUCESSFUL 0

/***** Serial Command Constants ******/
/* 	These constants coorespond to various
	PC->PIC and PIC->PC commands. If any
	byte is received on the serial port
	and it matches one of these constants,
	it is to be immediately treated as a
	command. If a data byte is send	that
	matches a command, it is escaped using
	the ESC command. Commands will only
	have a 0x00 - 0x7F range.
*/
// Shared commands
#define ESC 0x1b
#define ESC_MASK 0x80

// PC->PIC commands
#define BLOCK_UPLOAD 0x02
#define BLOCK_FINISH 0x03
#define BLOCK_DOWNLOAD 0x04
#define PING 0x05
#define SERVO_HOME 0x20
#define DISABLE_SERVO 0x21
#define ENABLE_SERVO 0x22
#define SNG_SERVO_CMD 0x23
#define ALL_SERVO_CMD 0x24
#define START_CTRL	  0x30
#define PAUSE_CTRL	  0x31
#define RESET_CTRL	  0x32	
#define SERVO_POS_REQUEST 0x40
#define FORCE_READ_REQUEST 0x41
#define SERVO_RAW_REQUEST 0x42
#define SENSOR_RAW_REQUEST 0x43
#define FORCE_TARE	0x44

// PIC->PC commands
#define PING_ECHO 0x07
#define ACK 0x06
#define NAK 0x15
#define SERVO_POS_PACKET 0x48
#define FORCE_READ_PACKET 0x49
#define SERVO_RAW_PACKET 0x4A
#define SENSOR_RAW_PACKET 0x4B

//Reset Constants
#define RESET_CMD 0x10
#define RESET_CONFIRM 0xFF
#define POWER_ON 0x0F
#define RCON_MASK 0b11100000
#define RCON_RESET 0b00010011

//Command states
#define NO_CMD_STATE 0
#define UPLOAD_CMD_STATE 1
#define RESET_CMD_STATE 2
#define SNG_SERVO_CMD_STATE 3
#define ALL_SERVO_CMD_STATE 4
#define DOWNLOAD_BLOCK 5
#define DOWLOAD_BLOCK_XMIT 6
#define FORCE_PACKET_CMD_STATE 7
static unsigned char command_state = 0;					//Current state of the serial communcations (i.e. if the serial port is awaiting the rest of a servo command or a block upload.)
static unsigned char force_packet_state = 1;
#define FORCE_PACKET_STATE_OFF		  0
#define FORCE_PACKET_STATE_CONTINUOUS 1
#define FORCE_PACKET_STATE_SINGLE     2

//Block upload/download
#define BLOCK_SIZE 128
#define TOTAL_BLOCK_BYTES (BLOCK_SIZE + 2)
#define BLOCK_OFFSET_INX 1
#define BLOCK_ID_INX 0
#define CHECKSUM_INX (TOTAL_BLOCK_BYTES - 1)
#define SRL_SUCCESSFUL 1
#define SRL_UNSUCCESSFUL 0
#define XMIT_BYTES_INTO_BUFFER 20
static unsigned char bytes_inx = 0;						//Number of bytes received during block transfers
static unsigned char block_buffer[TOTAL_BLOCK_BYTES];	//The buffer used between the communication line and the FLASH
static unsigned char xmit_checksum = 0;

//Servo Commands
static Leg_t LegNumber = NO_LEG;		//Flag to keep track of the leg of the servo being addressed
static ServoType_t	ServoType = NO_SRV; //Flag to keep track of the servo type of the servo being addresed
static ServoNumber_t ServoValues;		//Buffer used for a All_Servo_Cmd (so that the servos are updated at once.)

//Private function prototypes.
static void BlockData(unsigned char data);
static void BlockStart(void);
static unsigned char BlockEnd(void);
static void SngServoCmdStart(void);
static void SngServoCmdData(unsigned char data);
static void AllServoCmdStart(void);
static void AllServoCmdData(unsigned char data);
static void BlockXmitStart(unsigned char blockID);
static void BlockXmit(void);

//Private flags
static unsigned char escaped = 0;			//This flag keeps track if the last incoming byte was an escape byte.

//Transmit buffer
#define XMIT_BUFFER_SIZE 64
static unsigned char xmit_fifo_buffer[XMIT_BUFFER_SIZE];
static unsigned char xmit_pop_inx = 0;		//Index of the output of the FIFO
static unsigned char xmit_push_inx = 0;		//Index of the input of the FIFO (this index should point to an empty slot)

//Receive buffer
#define RCV_BUFFER_SIZE 8
static unsigned char rcv_fifo_buffer[RCV_BUFFER_SIZE];
static unsigned char rcv_pop_inx = 0;		//Index of the output of the FIFO
static unsigned char rcv_push_inx = 0;		//Index of the input of the FIFO (this index should point to an empty slot)


/************************ SRL_InitializeSerialPort **************************

Initializes the serial port and generate the power-on message.
This function will turn on the receive interrupt enable flag.

Last Modified
Salomon Trujillo		07-19-2006

*****************************************************************************/
void SRL_InitializeSerialPort(void)
{
	//Set Baudrate to 9600
	SPBRG = 64;

	//Setup PortC TX/RX pins
	TRISC &= ~(BIT6);	//TX
	TRISC |= (BIT7);	//RX

	//Enable Transmission, 8-bit, no-parity
	TXSTA = 0b00100000;

	//Enable Reception
	RCSTA = 0b10010000;

	//SendCmd power on command
	while(! PIR1bits.TXIF);
	TXREG = POWER_ON;

	//SendCmd RCON status
	while(! PIR1bits.TXIF);
	TXREG = RCON | RCON_MASK;

	//Reset RCON bits that must be software set.
	RCON |= RCON_RESET;

	//Enable Receive interrupts (transmit interrupt enable in SendCmd)
	PIE1bits.RCIE = 1;
}

/************************** SRL_HandleSerialPort ****************************

This function acts as the link between the software FIFOs and the serial port
hardware. This should be called by the interrupt service routine. This code
should be called by the interrupt service routine under one of the following
conditions:
1.) The reception register is full and must be emptied.
2.) The transmit register is empty and the transmit fifo has waiting data.

Last Modified
Salomon Trujillo		07-19-2006

*****************************************************************************/
void SRL_HandleSerialPort(void)
{
	//Check for an overrun error
	if (RCSTAbits.OERR)
	{
		//Need to reset the receive system, by cycling the enable bit
		RCSTAbits.CREN = 0;
		RCSTAbits.CREN = 1;
	}

	//Check for incoming data
	if (PIR1bits.RCIF)
	{
		//There is incoming data in the register. Retrieve it.
		rcv_fifo_buffer[rcv_push_inx] = RCREG;

		//Advance index
		rcv_push_inx = (rcv_push_inx + 1) % RCV_BUFFER_SIZE;
	}

	//Check for outgoing data
	if (xmit_pop_inx != xmit_push_inx)
	{
		//Check for an open buffer
		if (PIR1bits.TXIF)
		{
			//SendCmd data from buffer
			TXREG = xmit_fifo_buffer[xmit_pop_inx];
			
			//Advance index
			xmit_pop_inx = (xmit_pop_inx + 1) % XMIT_BUFFER_SIZE;
		}
	}

	//Note: I want this check done every time because
	//the program might have just finished a write to TXREG.
	//Using an else here instead of a second if will mean that
	//the interrupt will be triggered an extra time for each
	//block of tranmission
	if (xmit_pop_inx == xmit_push_inx)
	{
		//Buffer is empty, the transmit empty flag should
		//not trigger an interrupt, so disable it
		PIE1bits.TXIE = 0;
	}
}

/***************************** SRL_HandleData *******************************

This function handles all the data in the software FIFOs. It should be called
often, but it is not a high priority function. (Note: it should be called
faster than the 20ms control loop.) It acts as the interface between the FIFO
and the rest of the software. It will handle both incoming and outgoing data.
Furthermore, it will directly handle the simplest PC<->PIC commands.

Last Modified
Salomon Trujillo		09-06-2006

*****************************************************************************/
void SRL_HandleData(void)
{
	//Variables are static to increase code execution
	static unsigned char incoming_data;

	//Handle any outgoing data transmissions
	if (command_state == DOWLOAD_BLOCK_XMIT)
	{
		BlockXmit();
	}

	//Check for incoming data
	if (rcv_pop_inx != rcv_push_inx)
	{
		//There is incoming data in the register. Retrieve it.
		incoming_data = rcv_fifo_buffer[rcv_pop_inx];

		//Advance index
		rcv_pop_inx = (rcv_pop_inx + 1) % RCV_BUFFER_SIZE;

		//Check for escape byte
		if (incoming_data == ESC)
		{
			//Set the escape flag for the next byte.
			escaped = 1;
		}
		else
		{
			//Trigger commands
			switch (incoming_data)
			{
				case RESET_CMD:
					//Set state
					command_state = RESET_CMD_STATE;
				break;

				case BLOCK_UPLOAD:
					BlockStart();	
				break;
				
				case BLOCK_FINISH:
					//Finish the block transfer and respond
					if (BlockEnd() == SRL_SUCCESSFUL)
					{
						//A SRL_SUCCESSFUL upload. Send ACK
						SRL_SendCmd(ACK);
					}
					else
					{
						//An error occurred.
						SRL_SendCmd(NAK);
					}
				break;

				case BLOCK_DOWNLOAD:
					//Next byte will contain the block id. Start the transmit then.
					command_state = DOWNLOAD_BLOCK;
				break;

				case PING:
					//SendCmd Ping Echo
					SRL_SendCmd(PING_ECHO);
				break;
				
				//Servo Commands
				case SERVO_HOME:
					SRV_MoveToHome();
					SRL_SendCmd(ACK);
				break;

				case DISABLE_SERVO:
					SRV_DisableServos();
					SRL_SendCmd(ACK);
				break;

				case ENABLE_SERVO:
					SRV_EnableServos();
					SRL_SendCmd(ACK);
				break;

				case SNG_SERVO_CMD:
					SngServoCmdStart();
				break;

				case ALL_SERVO_CMD:
					AllServoCmdStart();
				break;

				//Controller commands
				case START_CTRL:
					CTRL_OL_EnableCtrl();
					CTRL_CL_EnableCtrl();
				break;

				case PAUSE_CTRL:
					CTRL_OL_EnableCtrl();
					CTRL_CL_DisableCtrl();
				break;

				case RESET_CTRL:
					CTRL_OL_EnableCtrl();
					CTRL_CL_ResetCtrl();
				break;

				//Report commands:
				case SERVO_POS_REQUEST:
					SRL_SendServos();
				break;
				
				case FORCE_READ_REQUEST:
					command_state = FORCE_PACKET_CMD_STATE;
				break;
				
				case SERVO_RAW_REQUEST:
					//SRL_SendRawServos();
				break;
				
				case FORCE_TARE:
					SNS_ZeroSensors();
				break;

				//No commands found, this must be a data byte
				default:
					if (escaped)
					{
						//The last byte was escape, so we need to modify the data and send it
						//To the data handler
						incoming_data = incoming_data ^ ESC_MASK;	
					}

					//SendCmd the data to the handler.
					switch(command_state)
					{
						case UPLOAD_CMD_STATE:
							BlockData(incoming_data);
						break;
						case SNG_SERVO_CMD_STATE:
							SngServoCmdData(incoming_data);
						break;
						case ALL_SERVO_CMD_STATE:
							AllServoCmdData(incoming_data);
						break;
						case RESET_CMD_STATE:
							//The reset function is small enough to include here.
							if (incoming_data == RESET_CONFIRM)
							{
								//Good confirmation. Reset the PIC
								_asm
									RESET
								_endasm
							}
							else
							{
								//Bad confirmation switch to old state
								command_state = NO_CMD_STATE;
							}
						break;
						case DOWNLOAD_BLOCK:
							//Start the download of a block from flash
							BlockXmitStart(incoming_data);	
						break;
						case FORCE_PACKET_CMD_STATE:
							force_packet_state = incoming_data;
							command_state = NO_CMD_STATE;
						break;
						default:
						break;
					}
				break;
			}
			//Since this byte wasn't escape, the next byte won't be escaped
			escaped = 0;
		}
	}
}

/****************************** SRL_SendCmd *********************************

This will send raw bytes without escaping them. This is good for sending
commands.

Last Modified
Salomon Trujillo		07-19-2006

*****************************************************************************/
void SRL_SendCmd(unsigned char data)
{
	//Save esc in buffer
	xmit_fifo_buffer[xmit_push_inx] = data;
	//Advance index
	xmit_push_inx = (xmit_push_inx + 1) % XMIT_BUFFER_SIZE;	

	//Enable transmit interrupts
	PIE1bits.TXIE = 1;
}

/****************************** SRL_SendData ********************************

This transmits a byte. If the byte matches a command byte it will escape it.
(See the TWiki for details.)

Last Modified
Salomon Trujillo		07-19-2006

*****************************************************************************/
void SRL_SendData(unsigned char data)
{
	//Check for data that needs to be escaped
	switch (data)
	{	
		case ACK:
		case NAK:
		case PING_ECHO:
		case POWER_ON:
		case ESC:
		case SERVO_POS_PACKET:
		case FORCE_READ_PACKET:
			//Modify data
			data = data ^ ESC_MASK;
			//SendCmd escape flag
			SRL_SendCmd(ESC);
			//Fall-through
		default:
			SRL_SendCmd(data);
	}
}

/******************************* BlockStart *********************************

This function is run once the PIC receives an upload to command to receive
flash data..


Last Modified
Salomon Trujillo		07-19-2006

*****************************************************************************/
static void BlockStart(void)
{
	//Reset buffer
	bytes_inx = 0;

	//Set state
	command_state = UPLOAD_CMD_STATE;
}

static void BlockData(unsigned char data)
{
	//Check for buffer overflow
	if (bytes_inx < TOTAL_BLOCK_BYTES)
	{
		block_buffer[bytes_inx] = data;
	}
	//Increment the number of bytes (even if we've received too many.)
	bytes_inx++;
}

//Returns whether or not the upload was SRL_SUCCESSFUL
static unsigned char BlockEnd(void)
{
	//Static for faster execution
	static unsigned char checksum;
	static unsigned char i;
	static unsigned short long RAMptr;

	//Check size
	if (bytes_inx != TOTAL_BLOCK_BYTES)
	{

		//Bad length
		return (SRL_UNSUCCESSFUL);
	}

	//Check the checksum
	checksum = 0;
	for(i = 0; i < TOTAL_BLOCK_BYTES; i++)
		checksum += block_buffer[i];

	if (checksum != 0)
	{
		//Bad checksum
		return (SRL_UNSUCCESSFUL);
	}

	//Find location in memory
	RAMptr = FLASH_DATA + (BLOCK_SIZE * block_buffer[BLOCK_ID_INX]);

	//Erase flash
	i = EraseFlash(RAMptr, BLOCK_SIZE);
	if (i != SUCCESS)
	{
		return(SRL_UNSUCCESSFUL);
	}

	//Write data
	i = WriteFlash(RAMptr, block_buffer + BLOCK_OFFSET_INX,
      BLOCK_SIZE);
	if (i != SUCCESS)
	{
		return(SRL_UNSUCCESSFUL);
	}

	//Everything worked
	return (SRL_SUCCESSFUL);	
}

static void BlockXmitStart(unsigned char blockID)
{
	static unsigned short long RAMptr;

	//Find location in memory
	RAMptr = FLASH_DATA + (BLOCK_SIZE * blockID);

	//Read flash
	ReadFlash(RAMptr, block_buffer, BLOCK_SIZE);	

	//Reset counter and checksum
	bytes_inx = 0;
	xmit_checksum = 0;

	//Setup transmission of bytes
	command_state = DOWLOAD_BLOCK_XMIT;
}

static void BlockXmit(void)
{
	static signed char bytes_avail;
	static unsigned char i;

	//We only want a maximum of XMIT_BYTES_INTO_BUFFER in the buffer after this BlockXmit (the constant assumes
	//as a worst case that all bytes must be escaped). To calculate the bytes_avail, we first find how many are pushed
	//into fifo by comparing push - pop (which must be taken modulo the buffer size.

	//bytes_avail = XMIT_BYTES_INTO_BUFFER - ((xmit_push_inx - xmit_pop_inx) &  XMIT_BUFFER_SIZE_PO2);

	//Transmit up to the bytes
	//for (i = 0; ((i < bytes_avail) && (bytes_inx < BLOCK_SIZE)); i++)
	
	if (xmit_push_inx == xmit_pop_inx)	
	{
		//Transmit the byte
		SRL_SendData(block_buffer[bytes_inx]);

		//Increment the checksum
		xmit_checksum += block_buffer[bytes_inx];

		//Increment counter
		bytes_inx++;
	}

	//Check for the end of the transmission
	if (bytes_inx == BLOCK_SIZE)
	{
		//Transmit checksum
		SRL_SendData(-xmit_checksum);

		//End transmission
		command_state = NO_CMD_STATE;
	}
}

static void SngServoCmdStart(void)
{
	//Set the servo index to wait for the first byte
	LegNumber = NO_LEG;
	ServoType = NO_SRV;

	//Set the state to servos
	command_state = SNG_SERVO_CMD_STATE;
}

static void SngServoCmdData(unsigned char data)
{
	static Number_t servo_target = 0;
	if (ServoType == NO_SRV)
	{
		//We must not have received any data, yet. This must be the servo index.
		//We also need to eventually change this to the new comm system (SINGLE_SERV_CMD, SRV_TYPE, LEG_NUM, VALUE (number?))
		ServoType = (data >> 2);
		LegNumber = (data & 0b00000011);
	}
	else if (LegNumber == NO_LEG)
	{
		// Fill in the leg number under the new system
	}
	else
	{
		servo_target.byParts.whole = data;
		
		#ifndef SRL_SRV_CMD_SND_TO_CTRL
			//We have already received the servo index. This must be the servo position.
			SRV_SetSingleAbs(LegNumber, ServoType, servo_target);
	
			//Commit the changes
			SRV_UpdateServoValues();
		#else
			SRV_SetOffset(LegNumber, ServoType, servo_target);
		#endif

		//Acknowledge the servo command
		SRL_SendCmd(ACK);
	}
}

static void AllServoCmdStart(void)
{
	/* Set the leg and servo type variables to indicate that we are starting
	 * on the first (FRONT_LEFT, FOOT_SRV) */
	LegNumber = 0;
	ServoType = 0;

	//Set the state to servos
	command_state = ALL_SERVO_CMD_STATE;
}

static void AllServoCmdData(unsigned char data)
{
	//Check for the number of data bytes received
	if ((LegNumber < NUM_LEGS) && (ServoType < NUM_SERVO_TYPES))
	{			
		/* Set the value (need to change if we go to Number_t representation */
		ServoValues.asTable[ServoType][LegNumber].byParts.whole = data;

		/* Increment the leg/servo numbers */
		LegNumber++;
		if (LegNumber == NUM_LEGS)
		{
			LegNumber = 0;
			ServoType++;
		}

		//Check to see if that's every servo
		if(ServoType == NUM_SERVO_TYPES)
		{
			//Set Targets
			SRV_SetAllAbs(&ServoValues);

			//Commit the changes
			SRV_UpdateServoValues();

			//Reply with ACK
			SRL_SendCmd(ACK);
		}
	}
	else
	{
		//Too many bytes have been received
		SRL_SendCmd(NAK);

		//Don't continue to send NAKs
		command_state = NO_CMD_STATE;		
	}
}

void SRL_SendServos(void)
{
	static Leg_t leg;
	static ServoType_t type;

	//Send packet command
	SRL_SendCmd(SERVO_POS_PACKET);

	//Cycle through each servo and send the whole-part of the target
	for (type = 0; type < NUM_SERVO_TYPES; type++)
		for (leg = 0; leg < NUM_LEGS; leg++)
			SRL_SendData(SRV_GetSingleTarget(leg, type).byParts.whole);
}

/*void SRL_SendRawServos(void)
{
	unsigned char i;
	ServoRaw_t RawServoValues;

	//Send packet command
	SRL_SendCmd(SERVO_RAW_PACKET);

	//Get the 
	SRV_GetAllRaw(&RawServoValues);

	//Cycle through each servo and send the whole-part of the target
	for (i = 0; i < NUM_SERVOS; i++)
		SRL_SendData(RawServoValues.asArray[i]);
}*/

void SRL_SendRawSensors(void)
{
	static Leg_t leg;
	static SensorType_t type;

	//Send packet command
	SRL_SendCmd(SENSOR_RAW_PACKET);

	//Cycle through each servo and send the whole-part of the target
	for (type = 0; type < NUM_SENSOR_TYPES; type++)
		for (leg = 0; leg < NUM_LEGS; leg++)
			SRL_SendData(SNS_GetSingleRaw(leg, type));
}

void SRL_SendForces(void)
{
	static Leg_t leg;
	static SensorType_t type;
	Number_t value;

	if (force_packet_state != FORCE_PACKET_STATE_OFF)
	{
		//Send packet command
		SRL_SendCmd(FORCE_READ_PACKET);
	
		//Cycle through each servo and send the whole-part of the target
		for (type = 0; type < NUM_SENSOR_TYPES; type++)
			for (leg = 0; leg < NUM_LEGS; leg++)
			{	
				//Retrieve force
				value = SNS_GetSingleForce(leg, type);
				
				//Send in little-endian format
				SRL_SendData(value.byParts.fraction);
				SRL_SendData(value.byParts.whole);
			}
		
		value.raw = Ctrl_OL_GetIndex();
		if (value.raw == 0xFFFF)
			value.raw = Ctrl_CL_GetIndex();
			
		SRL_SendData(value.byParts.fraction);
		SRL_SendData(value.byParts.whole);
		
		if (force_packet_state != FORCE_PACKET_STATE_CONTINUOUS)
			force_packet_state = FORCE_PACKET_STATE_OFF;
	}	
}

