#include "Servo.h"
#include "18f4520_Flash.h"
#include "FileReader.h"
#include <delays.h>
#include <p18cxxx.h>
#include "18fUtil.h"
#include "PhysicalConfig.h"
#include "ControllerMath.h"
#include "SBConfig.h"


/* Default limit values, incase there is no servo table entry */
#define DEFAULT_HOME		128
#define DEFAULT_UPPER		192
#define DEFAULT_LOWER		64


//ServoValues
static ServoRaw_t ServoValues;		//Actual position
static ServoNumber_t ServoTargets;	//Target position (with fractions)
//assume that LowerLimits<=HomePositions<=UpperLimits for all servos
static ServoRaw_t HomePositions;
static ServoRaw_t UpperLimits;
static ServoRaw_t LowerLimits;

/*FIX: Add this to the servo config table with home/limit positions, and read
 * in during Initialize_Servos().
 */
#ifdef VER_STICKYBOT
	static char ServoReversed[] = {0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0};
#endif	
#ifdef VER_STICKYBOT2
	static char ServoReversed[] = {1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0};
#endif

static ServoNumber_t ServoOffsets = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

//Shadow Registers
static unsigned char mServTempP1;
static unsigned char mServTempP2;
static unsigned char mServTempP3;

//Number of servo ticks
static unsigned char ServoTick;

static unsigned char servo_enabled;

//Private Functions
static void ServoLoop(void);
static void ReadServoTable(void);

void SRV_InitializeServos(void)
{
	///TEST
	unsigned char leg, type;
	
	//Ensure timer is disabled
	servo_enabled = 0;

	//Setup servo pins to outputs and low
	#if (NUM_OF_SERV_P1 > 0)
		TRIS_SERV_P1 = (~MASK_SERV_P1)&TRIS_SERV_P1;
		PORT_SERV_P1 = (~MASK_SERV_P1)&PORT_SERV_P1;
	#endif
	
	#if (NUM_OF_SERV_P2 > 0)
		TRIS_SERV_P2 = (~MASK_SERV_P2)&TRIS_SERV_P2;
		PORT_SERV_P2 = (~MASK_SERV_P2)&PORT_SERV_P2;
	#endif
	
	#if (NUM_OF_SERV_P3 > 0)
		TRIS_SERV_P3 = (~MASK_SERV_P3)&TRIS_SERV_P3;
		PORT_SERV_P3 = (~MASK_SERV_P3)&PORT_SERV_P3;
	#endif

	/* Fill the home and upper/lower limits with reasonable values in case
	 * there is no (or not enough) config data in flash. */
	for (leg = 0; leg < NUM_LEGS; leg++)
	{
		for (type = 0; type < NUM_SERVO_TYPES; type++)
		{
			HomePositions.asTable[type][leg] = DEFAULT_HOME;
			UpperLimits.asTable[type][leg] = DEFAULT_UPPER;
			LowerLimits.asTable[type][leg] = DEFAULT_LOWER;
		}
	}
	
	/* Now try to read from flash */
	ReadServoTable();

	//Initialize Servo position
	SRV_MoveToHome();
}

static void ReadServoTable()
{
	near unsigned short long data_loc;
	unsigned char num_entries;
	
	data_loc = (near unsigned short long)FR_FindEntry(SERVO_CONFIG_TABLE);
	
	/* If there is a servo configuration table, get the data */
	if (data_loc != NO_ENTRY_FOUND)
	{
		/* Only want to read up to the number of servos we have */
		num_entries = ReadFlashByte(data_loc);
		if (num_entries != NUM_SERVOS)
		{
			// DO Something else to indicate the mismatch?
			if (num_entries > NUM_SERVOS)
			{
				num_entries = NUM_SERVOS;
			}
		}
		data_loc++;
		
		/* Read in the table of home positions */
		ReadFlash(data_loc, &HomePositions, num_entries);
		data_loc += num_entries;
		
		/* Read in the table of upper limits */
		ReadFlash(data_loc, &UpperLimits, num_entries);
		data_loc += num_entries;
		
		/* Read in the table of lower limits */
		ReadFlash(data_loc, &LowerLimits, num_entries);
	}
}


//Reset all the servos to their home position
void SRV_MoveToHome(void)
{
	static unsigned char type;
	static unsigned char leg;

	/* Loop through to set all the targets to 0 */
	for (leg = 0; leg < NUM_LEGS; leg++)
	{
		for (type = 0; type < NUM_SERVO_TYPES; type++)
		{
			ServoTargets.asTable[type][leg].val = 0;
			ServoOffsets.asTable[type][leg].val = 0;
		}
	}

	/* Set the actual values */
	ServoValues = HomePositions;
}

//-----------------------------------------------
//	(2 * NUM_OF_SERV) + 5 instructions	= 29 (including call and return)
static void ServoLoop(void)
{
	/* Looks at which tick number we're on (0-255) and compares that value to
	 * the servo value for each servo.  If the servo value is greater the
	 * pwm for that servo needs to be high longer, so the pin stays high,
	 * otherwise the pin is cleared.
	 *
	 * CPFSGT - skips next instruction if file value is greater than W
	 */	
	_asm 
		movf ServoTick,0,1 
		#if (NUM_OF_SERV_P1 > 0) 
			CPFSGT	ServoValues,1 
			BCF	mServTempP1,0,1 
		#endif 
		#if (NUM_OF_SERV_P1 > 1) 
			CPFSGT	ServoValues+1,1 
			BCF	mServTempP1,1,1
		#endif
		#if (NUM_OF_SERV_P1 > 2)
			CPFSGT	ServoValues+2,1
			BCF	mServTempP1,2,1
		#endif
		#if (NUM_OF_SERV_P1 > 3)
			CPFSGT	ServoValues+3,1
			BCF	mServTempP1,3,1
		#endif
		#if (NUM_OF_SERV_P1 > 4)
			CPFSGT	ServoValues+4,1
			BCF	mServTempP1,4,1
		#endif
		#if (NUM_OF_SERV_P1 > 5) 
			CPFSGT	ServoValues+5,1
			BCF	mServTempP1,5,1
		#endif
		#if (NUM_OF_SERV_P1 > 6)
			CPFSGT	ServoValues+6,1
			BCF	mServTempP1,6,1
		#endif
		#if (NUM_OF_SERV_P1 > 7)
			CPFSGT	ServoValues+7,1
			BCF	mServTempP1,7,1
		#endif
		#if (NUM_OF_SERV_P2 > 0) 
			CPFSGT	ServoValues+8,1
			BCF	mServTempP2,0,1
		#endif
		#if (NUM_OF_SERV_P2 > 1) 
			CPFSGT	ServoValues+9,1
			BCF	mServTempP2,1,1
		#endif
		#if (NUM_OF_SERV_P2 > 2) 
			CPFSGT	ServoValues+10,1
			BCF	mServTempP2,2,1
		#endif
		#if (NUM_OF_SERV_P2 > 3) 
			CPFSGT	ServoValues+11,1
			BCF	mServTempP2,3,1
		#endif
		#if (NUM_OF_SERV_P2 > 4)
			CPFSGT	ServoValues+12,1
			BCF	mServTempP2,4,1
		#endif
		#if (NUM_OF_SERV_P2 > 5)
			CPFSGT	ServoValues+13,1
			BCF	mServTempP2,5,1
		#endif
		#if (NUM_OF_SERV_P2 > 6)
			CPFSGT	ServoValues+14,1
			BCF	mServTempP2,6,1
		#endif
		#if (NUM_OF_SERV_P2 > 7)
			CPFSGT	ServoValues+15,1
			BCF	mServTempP2,7,1
		#endif
		#if (NUM_OF_SERV_P3 > 0)
			CPFSGT	ServoValues+16,1
			BCF	mServTempP3,0,1
		#endif
		#if (NUM_OF_SERV_P3 > 1)
			CPFSGT	ServoValues+17,1
			BCF	mServTempP3,1,1
		#endif
		#if (NUM_OF_SERV_P3 > 2) 
			CPFSGT	ServoValues+18,1
			BCF	mServTempP3,2,1
		#endif 
		#if (NUM_OF_SERV_P3 > 3)
			CPFSGT	ServoValues+19,1 
			BCF	mServTempP3,3,1 
		#endif
		#if (NUM_OF_SERV_P3 > 4)
			CPFSGT	ServoValues+20,1
			BCF	mServTempP3,4,1
		#endif
		#if (NUM_OF_SERV_P3 > 5) 
			CPFSGT	ServoValues+21,1 
			BCF	mServTempP3,5,1
		#endif
		#if (NUM_OF_SERV_P3 > 6)  
			CPFSGT	ServoValues+22,1 
			BCF	mServTempP3,6,1
		#endif 
		#if (NUM_OF_SERV_P3 > 7)
			CPFSGT	ServoValues+23,1
			BCF	mServTempP3,7,1
		#endif
	_endasm
}


//---------------------------------------------------------------------------------------//

void SRV_GenerateServoPWM(void)
{
	if (servo_enabled)
	{
		ServoTick = 0;
	
		#if (NUM_OF_SERV_P1 > 0)
			mServTempP1 = 0xFF;
			PORT_SERV_P1 = PORT_SERV_P1 | MASK_SERV_P1;
		#endif
		
		#if (NUM_OF_SERV_P2 > 0)
			mServTempP2 = 0xFF;
			PORT_SERV_P2 = PORT_SERV_P2 | MASK_SERV_P2;
		#endif
		
		#if (NUM_OF_SERV_P3 > 0)
			mServTempP3 = 0xFF;
			PORT_SERV_P3 = PORT_SERV_P3 | MASK_SERV_P3;
		#endif
			
		// DelayMicroSec	D'860'	// 0.860 ms DI ATTESA (per utilizzare tutti i bit ... 
					// ... vado leggermente fuori le specifiche del servo)
		Delay100TCYx(80);	// 800 us
		Delay10TCYx(60);	// 60 us
				
		do   // ServoLoop should run in exactly 5 us!!
		{
			// IL LOOP, ESCLUSI I NOP, E' DI (2 * NUM_OF_SERV + 2 * NUM_SERVOPORT + 4) ISTRUZIONI <<--VERIFY!!
			
			ServoLoop();			// La macro occupa (2 * NUM_OF_SERV) Istruzioni.
			//N_Nop USE A HW TIMER?? - Sim shows exactly 5us.	
			Nop();
			Nop();
			Nop();
			Nop();
			Nop();
			Nop();
			Nop();	// Determined via simulation
								
			#if (NUM_OF_SERV_P1 > 0)
				PORT_SERV_P1 = PORT_SERV_P1 & mServTempP1;
			#endif
			#if (NUM_OF_SERV_P2 > 0)
				PORT_SERV_P2 = PORT_SERV_P2 & mServTempP2;
			#endif
			#if (NUM_OF_SERV_P3 > 0)
				PORT_SERV_P3 = PORT_SERV_P3 & mServTempP3;
			#endif
			ServoTick++;
		} while (ServoTick != 0);
	}
}

void SRV_EnableServos(void)
{
	servo_enabled = 1;
}

void SRV_DisableServos(void)
{
	servo_enabled = 0;
}

void SRV_ToggleServos(void)
{
	servo_enabled = !servo_enabled;
}

/* Commits the current values in ServoTargets to ServoValues (which are used
 * when generating the PWM).  Limits are checked here, and the target values
 * may be changed to prevent wind-up effects.
 */
void SRV_UpdateServoValues()
{
	static unsigned char type;
	static unsigned char leg;
	static unsigned char reversed;
	signed int val;
	
	for (leg = 0; leg < NUM_LEGS; leg++)
	{
		for (type = 0; type < NUM_SERVO_TYPES; type++)
		{
			/* Get the whole number value from the servo targets and add it to
			 * the home position (since the target is relative to home).  Also
			 * handle switching the sign for certain motors.
			 */
			val = HomePositions.asTable[type][leg];

			/* Set reversed to 0 if the servo is reversed (ie. + => down) */
			reversed = ServoReversed[NUM_LEGS*type + leg];
			
			if (reversed)
			{
				val -= (signed int)ServoTargets.asTable[type][leg].byParts.whole;
			}
			else
			{
				val += (signed int)ServoTargets.asTable[type][leg].byParts.whole;
			}
			
			//Check the upper limit
			if (val > UpperLimits.asTable[type][leg])
			{
				/* Set the servo value to the upper limit */
				ServoValues.asTable[type][leg] = UpperLimits.asTable[type][leg];
				
				/* Reset the target; anti wind-up */
				if (reversed)
				{
					ServoTargets.asTable[type][leg].byParts.whole = \
						HomePositions.asTable[type][leg] - UpperLimits.asTable[type][leg];
				}
				else
				{
					ServoTargets.asTable[type][leg].byParts.whole = \
						UpperLimits.asTable[type][leg] - HomePositions.asTable[type][leg];
				}
				ServoTargets.asTable[type][leg].byParts.fraction = 0;
			}
			//Check the lower limit
			else if (val < LowerLimits.asTable[type][leg])
			{
				/* Set the servo value to the lower limit */
				ServoValues.asTable[type][leg] = LowerLimits.asTable[type][leg];
				
				/* Reset the target; anti wind-up */
				if (reversed)
				{
					ServoTargets.asTable[type][leg].byParts.whole = \
						HomePositions.asTable[type][leg] - LowerLimits.asTable[type][leg];
				}
				else
				{
					ServoTargets.asTable[type][leg].byParts.whole = \
						LowerLimits.asTable[type][leg] - HomePositions.asTable[type][leg];
				}
				ServoTargets.asTable[type][leg].byParts.fraction = 0;
			}
			//The servo is in bounds.
			else
			{
				ServoValues.asTable[type][leg] = val;
			}
		}
	}
}

void SRV_SetSingleAbs(Leg_t leg, ServoType_t type, Number_t position)
{
	ServoTargets.asTable[type][leg] = position;
}

	
void SRV_SetSingleRel(Leg_t leg, ServoType_t type, Number_t position)
{
	ServoTargets.asTable[type][leg].raw += position.raw;
}
		
	
void SRV_SetGroupAbs(ServoType_t type, ServoGroupNumber_t *positions)
{
	ServoTargets.asTypeArray[type] = *positions;
}


void SRV_SetGroupRel(ServoType_t type, ServoGroupNumber_t *positions)
{
	unsigned char leg;
	NumberSum_t sum;
	
	for (leg = 0; leg < NUM_LEGS; leg++)
		ServoTargets.asTable[type][leg].raw += positions->asValArray[leg].raw;
}


void SRV_SetAllAbs(ServoNumber_t *positions)
{
	ServoTargets = *positions;
}


void SRV_SetAllRel(ServoNumber_t *positions)
{
	static unsigned char leg;
	static unsigned char type;
	static NumberSum_t sum;
	
	for (leg = 0; leg < NUM_LEGS; leg++)
		for (type = 0; type < NUM_SERVO_TYPES; type++)
			ServoTargets.asTable[type][leg].raw += positions->asTable[type][leg].raw;
}

Number_t SRV_GetSingleTarget(Leg_t leg, ServoType_t type)
{
	return(ServoTargets.asTable[type][leg]);
}


void SRV_GetGroupTargets(ServoType_t type, ServoGroupNumber_t *dest)
{
	*dest = ServoTargets.asTypeArray[type];
}

void SRV_GetAllTargets(ServoNumber_t *dest)
{
	*dest = ServoTargets;
}

void SRV_SetOffset(Leg_t leg, ServoType_t type, Number_t offset)
{
	ServoOffsets.asTable[type][leg] = offset;
}

ServoNumber_t * SRV_GetOffset(void)
{
	return &ServoOffsets;
}
